/*
 * Copyright (C) 1996-2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS$");

/**
 ** Commandes GL disponibles en TCL
 **/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdint.h>

#include <tcl.h>
#include <togl.h>

#include "global.h"
#include "commands.h"
#include "cb.h"

/* Compteur des display lists */
static int gdheNList = 0;

/*----------------------------------------------------------------------*/

/**
 ** definit les parametres de la 'viewing transformation
 **/
int
setObsCallBack(struct Togl *togl, int argc, char *argv[])
{
    Tcl_Interp *interp = Togl_Interp(togl);
    Tcl_HashEntry *entry;
    GDHE_OBS *obs;
    Tcl_Obj *result[7];
    int i,n;

    /* recherche dans la H-table */
    entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
    if (entry == NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("unknown togl ident", -1));
	return TCL_ERROR;
    }
    obs = (GDHE_OBS *)Tcl_GetHashValue(entry);

    for (i = 2; i < argc; i++) {
        /* la situation de l'observateur */
	if (strcmp(argv[i], "-dist") == 0) {
	    obs->dist = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-elev") == 0) {
	    obs->elev = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-azi") == 0) {
	    obs->azi = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-x") == 0) {
	    obs->x = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-y") == 0) {
	    obs->y = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-z") == 0) {
	    obs->z = atof(argv[++i]);
	    continue;
	}
	if (strcmp(argv[i], "-euler") == 0) {
		/* undocumented */
		obs->viewMatMode = 2;
		continue;
	}
	if (strcmp(argv[i], "-classic") == 0) {
		/* Reset "classic" gdhe observer positionning mode */
		obs->viewMatMode = 0;
		continue;
	}
	if (strcmp(argv[i], "-viewMat") == 0) {
	  n = 0;
	  while ( (i+1 < argc ) && (n<16) &&
		  ( !isnan( obs->viewMat[n] = atof(argv[++i]) )) ) {
	    n++;
	  }
	  if ( n==16 ) {
	    obs->viewMatMode = 1;
	    continue;
	  } else {
	    printf("not enough values in the viewer matrix\n");
	  }
	}

	/* les modes de projection */
	if (strcmp(argv[i], "-perspective") == 0) {
	    obs->projMode = PERSPECTIVE;
	    continue;
	}
	if (strcmp(argv[i], "-ortho") == 0) {
	     obs->projMode = ORTHOGONAL;
	     continue;
	}
	if (strcmp(argv[i], "-projMat") == 0) {
	  n = 0;
	  while ( (i+1 < argc ) && (n<16) &&
		  ( !isnan( obs->projMat[n] = atof(argv[++i]) )) ) {
	    n++;
	  }
	  if ( n==16 ) {
	    obs->projMode = CUSTOM;
	    continue;
	  } else {
	    printf("not enough values in the projection matrix\n");
	  }
	}

	/* les parametres de projection */
	if (strcmp(argv[i], "-fovy") == 0) {
	    obs->fovy = atof(argv[++i]);
	    if (obs->fovy < 0.0) {
		obs->fovy = 0.0;
	    }
	    if (obs->fovy > 180.0) {
		obs->fovy = 180.0;
	    }
	    continue;
	}

	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("unknown arg: should be"
					  "\"path setObs -dist dist -elev elev -azi azi -viewMat ei"
					  " -x x -y y -z z -perspective|-ortho|-projMat\"",
					  -1));
	return TCL_ERROR;
    } /* for */

    result[0] = Tcl_NewDoubleObj(obs->dist);
    result[1] = Tcl_NewDoubleObj(obs->elev);
    result[2] = Tcl_NewDoubleObj(obs->azi);
    result[3] = Tcl_NewDoubleObj(obs->x);
    result[4] = Tcl_NewDoubleObj(obs->y);
    result[5] = Tcl_NewDoubleObj(obs->z);
    result[6] = Tcl_NewDoubleObj(obs->fovy);

    Tcl_SetObjResult(interp, Tcl_ConcatObj(7, result));
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

int
redrawCallBack(struct Togl *togl, int argc, char *argv[])
{
    Tcl_Interp *interp = Togl_Interp(togl);
    int bool = 0;

    if (argc != 2 && argc != 3) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
						  "\"widget redraw ?force?\"",
						  -1));
	return TCL_ERROR;
    }
    if (argc == 3) {
	if (Tcl_GetBoolean(interp, argv[2], &bool) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if (bool) {
	/* Raffichage immdiat */
	Togl_MakeCurrent(togl);
/* 	reshape_cb(togl); */
	display_cb(togl);
    } else {
 	Togl_PostRedisplay(togl);
    }

    /* Let result string empty */
    interp->result[0] = '\0';
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/*
 * Copie d'ecran PostScript
 */
int
dumpEpsCallBack(struct Togl *togl, int argc, char *argv[])
{
    Tcl_Interp *interp = Togl_Interp(togl);
    Tcl_Obj *usage =
	Tcl_NewStringObj("usage: \"widget dumpEps ?-color? file\"", -1);
    int color = 0;
    char *filename;

    switch (argc) {
      case 3:
	filename = argv[2];
	break;
      case 4:
	if (!strcmp(argv[2], "-color")) {
	    color = 1;
	    filename = argv[3];
	} else {
	    Tcl_SetObjResult(interp, usage);
	    return TCL_ERROR;
	}
	break;
      default:
	Tcl_SetObjResult(interp, usage);
	return TCL_ERROR;
    } /* switch */
    if (Togl_DumpToEpsFile(togl, filename, color, display_cb) != 0) {
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("Error in Togl_DumpToEpsFile", -1));
	return TCL_ERROR;
    }
    interp->result[0] = '\0';
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/*
 * Copie d'ecran PGM/PPM
 */
int
dumpPpmCallBack(struct Togl *togl, int argc, char *argv[])
{
    Tcl_Interp *interp = Togl_Interp(togl);
    Tcl_Obj *usage =
	Tcl_NewStringObj("usage: \"widget dumpPpm ?-color? file\"", -1);
    int color = 0;
    char *filename;

    switch (argc) {
      case 3:
	filename = argv[2];
	break;
      case 4:
	if (!strcmp(argv[2], "-color")) {
	    color = 1;
	    filename = argv[3];
	} else {
	    Tcl_SetObjResult(interp, usage);
	    return TCL_ERROR;
	}
	break;
      default:
	Tcl_SetObjResult(interp, usage);
	return TCL_ERROR;
    } /* switch */
    switch (Togl_DumpToPpmFile(togl, filename, color, display_cb)) {
      case 0:
	interp->result[0] = '\0';
	return TCL_OK;
      case 1:
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("malloc error in "
					  "Togl_DumpToPpmFile", -1));
	return TCL_ERROR;
      case 2:
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("write error in "
					  "Togl_DumpToPpmFile", -1));
	return TCL_ERROR;
      default:
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("unknown error in "
					  "Togl_DumpToPpmFile", -1));
	return TCL_ERROR;

    } /* switch */
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glColor'
 **/
int
colorCallBack(ClientData clientData, Tcl_Interp *interp,
	      int argc, Tcl_Obj *const argv[])
{
    int r, g, b, a;

    if (argc != 4 && argc != 5) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "color r g b ?a?\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &r);
    Tcl_GetIntFromObj(interp, argv[2], &g);
    Tcl_GetIntFromObj(interp, argv[3], &b);
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    if (argc == 5) {
	Tcl_GetIntFromObj(interp, argv[4], &a);
	glColor4ub(r, g, b, a);
    } else {
	glColor3ub(r, g, b);
    }

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glClearColor'
 **/
int
clearColorCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    double tmp;

    if (argc != 4 && argc != 5) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "clearColor r g b ?a? \"",
						  -1));
	return TCL_ERROR;
    }
    /* Initialise une variable globale utilise'e dans redrawCallBack */
    Tcl_GetDoubleFromObj(interp, argv[1], &tmp);
    backgroundColor[0] = tmp/255.0;
    Tcl_GetDoubleFromObj(interp, argv[2], &tmp);
    backgroundColor[1] = tmp/255.0;
    Tcl_GetDoubleFromObj(interp, argv[3], &tmp);
    backgroundColor[2] = tmp/255.0;
    if (argc == 5) {
	Tcl_GetDoubleFromObj(interp, argv[3], &tmp);
	backgroundColor[3] = tmp/255.0;
    } else {
	backgroundColor[3] = 1.0;
    }
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glMaterial'
 **/
int
materialCallBack(ClientData clientData, Tcl_Interp *interp,
	      int argc, Tcl_Obj *const argv[])
{
    GLfloat mat[4];
    double x;
    int op, narg;
    char *s;

    if (argc < 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "material type ....\"",
						  -1));
	return TCL_ERROR;
    }
    s = Tcl_GetStringFromObj(argv[1], NULL);
    if (!strcmp(s, "specular")) {
	op = GL_SPECULAR;
	narg = 4;
    } else if (!strcmp(s, "shininess")) {
	op = GL_SHININESS;
	narg = 1;
    } else if (!strcmp(s, "emission")) {
	op = GL_EMISSION;
	narg = 4;
    } else if (!strcmp(s, "ambient")) {
	op = GL_AMBIENT;
	narg = 4;
    } else if (!strcmp(s, "diffuse")) {
	op = GL_DIFFUSE;
	narg = 4;
    } else {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("syntax error: should be \""
						  "material <op> ...\"",
						  -1));
	return TCL_ERROR;
    }

    if (argc != 2+narg) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "material <op> r g b a\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[2], &x);
    mat[0] = (GLfloat)x;
    if (narg > 1) {
	Tcl_GetDoubleFromObj(interp, argv[3], &x);
	mat[1] = (GLfloat)x;
	Tcl_GetDoubleFromObj(interp, argv[4], &x);
	mat[2] = (GLfloat)x;
	Tcl_GetDoubleFromObj(interp, argv[5], &x);
	mat[3] = (GLfloat)x;
    }
    glDisable(GL_COLOR_MATERIAL);
    glMaterialfv(GL_FRONT, op, mat);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Display a string
 **/
int
drawStringCallBack(ClientData clientData, Tcl_Interp *interp,
		int argc, Tcl_Obj *const argv[])
{
     double x,y,z;
     char *s;

     if ((argc != 5)) {
	  Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						    "drawString x y z string \"",
						    -1));
	  return TCL_ERROR;
     }
     Tcl_GetDoubleFromObj(interp, argv[1], &x);
     Tcl_GetDoubleFromObj(interp, argv[2], &y);
     Tcl_GetDoubleFromObj(interp, argv[3], &z);
     s = Tcl_GetStringFromObj(argv[4], NULL);
     glDisable(GL_DEPTH_TEST);
     glDisable(GL_LIGHTING);
     glRasterPos3d(x,y,z);
     glListBase(FontBase);
     glCallLists(strlen(s), GL_UNSIGNED_BYTE, s);
     glEnable(GL_LIGHTING);
     glEnable(GL_DEPTH_TEST);

     /* Let result string empty */
     Tcl_SetObjResult(interp, nullObj);
     return TCL_OK;
}

/*----------------------------------------------------------------------*/

int
frameAnnotateCallBack(ClientData clientData, Tcl_Interp *interp,
		int argc, Tcl_Obj *const argv[])
{
	char *s;

	if (argc != 4) {
		Tcl_SetObjResult(interp,
		    Tcl_NewStringObj("wrong # args: should be "
			"frameAnnotate x y text", -1));
		return TCL_ERROR;
	}
	Tcl_GetIntFromObj(interp, argv[1], &annotateX);
	Tcl_GetIntFromObj(interp, argv[2], &annotateY);
	s = Tcl_GetStringFromObj(argv[3], NULL);
	annotateText = strdup(s);
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glTranslate'
 **/
int
translateCallBack(ClientData clientData, Tcl_Interp *interp,
		  int argc, Tcl_Obj *const argv[])
{
    double x, y, z;

    if (argc != 4) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "translate ?x? ?y? ?z? \"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &x);
    Tcl_GetDoubleFromObj(interp, argv[2], &y);
    Tcl_GetDoubleFromObj(interp, argv[3], &z);
    glTranslated(x, y, z);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glRotate'
 **/
int
rotateCallBack(ClientData clientData, Tcl_Interp *interp,
	       int argc, Tcl_Obj *const argv[])
{
    double a, x, y, z;

    if (argc != 5) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "rotate angle x y z \"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetDoubleFromObj(interp, argv[1], &a);
    Tcl_GetDoubleFromObj(interp, argv[2], &x);
    Tcl_GetDoubleFromObj(interp, argv[3], &y);
    Tcl_GetDoubleFromObj(interp, argv[4], &z);
    glRotated(a, x, y, z);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glPushMatrix'
 **/
int
pushMatrixCallBack(ClientData clientData, Tcl_Interp *interp,
		   int argc, Tcl_Obj *const argv[])
{

    if (argc != 1) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "pushMatrix \"",
						  -1));
	return TCL_ERROR;
    }
    glPushMatrix();
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glPopMatrix'
 **/
int
popMatrixCallBack(ClientData clientData, Tcl_Interp *interp,
		  int argc, Tcl_Obj *const argv[])
{

    if (argc != 1) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "popMatrix  \"",
						  -1));
	return TCL_ERROR;
    }
    glPopMatrix();
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glLoadIdentity'
 **/
int
loadIdentityCallBack(ClientData clientData, Tcl_Interp *interp,
		     int argc, Tcl_Obj *const argv[])
{

    if (argc != 1) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "loadIdentity  \"",
						  -1));
	return TCL_ERROR;
    }
    glLoadIdentity();
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glLoadMatrix'
 **/
int
loadMatrixCallBack(ClientData clientData, Tcl_Interp *interp,
		  int argc, Tcl_Obj *const argv[])
{
     double tmp;
     int i;
     GLdouble m[16];
     if (argc != 17) {
       Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						 "loadMatrix e00 e10 ... e23 e33 \"",
						 -1));
       return TCL_ERROR;
     }

     for (i = 1; i < 17;i++) {
	  Tcl_GetDoubleFromObj(interp, argv[i], &tmp);
	  m[i-1] = (GLdouble)tmp;
     }

     glLoadMatrixd(m);
     if ( glGetError() == GL_INVALID_OPERATION ) {
       printf("pb\n");
     }
     /* Let result string empty */
     Tcl_SetObjResult(interp, nullObj);
     return TCL_OK;
}
/**
 ** interface de 'getMVMatrix'
 **/
int
getMVMatrixCallBack(ClientData clientData, Tcl_Interp *interp,
		  int argc, Tcl_Obj *const argv[])
{
    Tcl_Obj *result[16];
    GLdouble m[16];
    int i;
    if (argc != 1) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
					      "getMVMatrix \"",
					      -1));
    return TCL_ERROR;
    }
    glGetDoublev(GL_MODELVIEW_MATRIX,m);

    if ( glGetError() == GL_INVALID_OPERATION ) {
      printf("pb\n");
    }
    for (i=0;i<16;i++) {
      result[i] = Tcl_NewDoubleObj(m[i]);
    }
    Tcl_SetObjResult(interp, Tcl_ConcatObj(16,result));
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'glMultMatrix'
 **/
int
multMatrixCallBack(ClientData clientData, Tcl_Interp *interp,
		  int argc, Tcl_Obj *const argv[])
{
     double tmp;
     int i;
     GLdouble m[16];
     if (argc != 17) {
 	  Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						    "multMatrix e00 e10 ... e23 e33 \"",
						    -1));
	  return TCL_ERROR;
     }

     for (i = 1; i < 17;i++) {
	  Tcl_GetDoubleFromObj(interp, argv[i], &tmp);
	  m[i-1] = (GLdouble)tmp;
     }

     glMultMatrixd(m);
     if ( glGetError() == GL_INVALID_OPERATION ) {
       printf("pb\n");
     }
     /* Let result string empty */
     Tcl_SetObjResult(interp, nullObj);
     return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glGenLists'
 **/
int
genListsCallBack(ClientData clientData, Tcl_Interp *interp,
		 int argc, Tcl_Obj *const argv[])
{
    int indice, n;

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "genLists n\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &n);
    indice = glGenLists(n);

    Tcl_SetObjResult(interp, Tcl_NewIntObj(indice));
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'glDeleteLists'
 **/
int
deleteListsCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    int indice, n;

    if (argc != 3 && argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "deleteLists indice n\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &indice);
    if (argc == 3) {
	Tcl_GetIntFromObj(interp, argv[2], &n);
    } else {
	n = 1;
    }
    glDeleteLists(indice, n);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glCallList'
 **/
int
callListCallBack(ClientData clientData, Tcl_Interp *interp,
		 int argc, Tcl_Obj *const argv[])
{
    int n;

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "callList n\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &n);
    glCallList(n);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'glNewList'
 **/
int
newListCallBack(ClientData clientData, Tcl_Interp *interp,
		int argc, Tcl_Obj *const argv[])
{
    int indice;

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "newList indice \"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &indice);
    glNewList(indice, GL_COMPILE_AND_EXECUTE);
    gdheNList++;
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'glEndList'
 **/
int
endListCallBack(ClientData clientData, Tcl_Interp *interp,
		int argc, Tcl_Obj *const argv[])
{

    if (argc != 1) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "endList\"",
						  -1));
	return TCL_ERROR;
    }
    glEndList();

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'glCullFace'
 **/
int
cullFaceCallBack(ClientData clientData, Tcl_Interp *interp,
		 int argc, Tcl_Obj *const argv[])
{
    int bool;

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "cullFace bool \"",
						  -1));
	return TCL_ERROR;
    }
    if (Tcl_GetBooleanFromObj(interp, argv[1], &bool) != TCL_OK) {
	return TCL_ERROR;
    }
    if (bool) {
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
    } else {
	glDisable(GL_CULL_FACE);
    }
    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** interface de 'glPolygonMode'
 **/
int
polygonModeCallBack(struct Togl *togl, int argc, char *argv[])
{

    Tcl_Interp *interp = Togl_Interp(togl);

    if (argc != 3) {
	Tcl_SetResult(interp, "wrong # args: should be \""
		      "widget polygonMode line|fill \"",
		      TCL_STATIC);
	return TCL_ERROR;
    }
    Togl_MakeCurrent(togl);
    if (strcmp(argv[2], "fill") == 0) {
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    } else if (strcmp(argv[2], "line") == 0) {
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    } else {
	Tcl_SetObjResult(interp,
			 Tcl_NewStringObj("unrecognized arg: should be "
					  "\"line\" or \"fill\"", -1));
	return TCL_ERROR;
    }

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** interface de 'Tcl_Sleep'
 **/
int
sleepCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    int millis;

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
						  "sleep miliseconds\"",
						  -1));
	return TCL_ERROR;
    }
    Tcl_GetIntFromObj(interp, argv[1], &millis);
    Tcl_Sleep(millis);

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/
static unsigned int gdheNameCounter = 0;
#define BUFSIZE 512
static Tcl_Obj* revNamesArray[BUFSIZE];

static void
gdheInitNames(Tcl_Interp *interp)
{
	int i;

	for (i = 0; i < gdheNameCounter; i++) {
		if (revNamesArray[i] != NULL) {
			Tcl_DecrRefCount(revNamesArray[i]);
			revNamesArray[i] = NULL;
		}
	}
	gdheNameCounter = 0;
	Tcl_UnsetVar(interp, "gdheNameStack", TCL_GLOBAL_ONLY);
	glInitNames();
}

/*----------------------------------------------------------------------*/

static unsigned int
gdheGetNameFromObj(Tcl_Interp *interp, Tcl_Obj *index)
{
	unsigned int i;
	Tcl_Obj *namesArray, *name, *iname;

	namesArray = Tcl_NewStringObj("gdheNameStack", -1);
	name = Tcl_ObjGetVar2(interp, namesArray, index, TCL_GLOBAL_ONLY);
	if (name != NULL) {
		Tcl_GetIntFromObj(interp, name, (int *)&i);
	} else {
		i = gdheNameCounter++;
		iname = Tcl_NewIntObj(i);
		Tcl_ObjSetVar2(interp, namesArray, index, iname,
		    TCL_GLOBAL_ONLY);
		revNamesArray[i] = index;
		Tcl_IncrRefCount(index);
	}
	return i;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glInitNames'
 **/
int
initNamesCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{

	if (argc != 1) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("wrong # args: should be \""
			"initNames \"", -1));
		return TCL_ERROR;
	}
	gdheInitNames(interp);
	
	/* Let result string empty */
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glPopName'
 **/
int
popNameCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{

	if (argc != 1) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("wrong # args: should be \""
			"popName \"", -1));
		return TCL_ERROR;
	}
	glPopName();
	
	/* Let result string empty */
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
}


/*----------------------------------------------------------------------*/
/**
 ** interface de 'glPushName'
 **/
int
pushNameCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
	if (argc != 2) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("wrong # args: should be \""
			"pushName name \"", -1));
		return TCL_ERROR;
	}
	glPushName(gdheGetNameFromObj(interp, argv[1]));

	/* Let result string empty */
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
}

/*----------------------------------------------------------------------*/
/**
 ** interface de 'glLoadName'
 **/
int
loadNameCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{

	if (argc != 2) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("wrong # args: should be \""
			"loadName name\"", -1));
		return TCL_ERROR;
	}
	glLoadName(gdheGetNameFromObj(interp, argv[1]));
	    
	/* Let result string empty */
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
}

/*----------------------------------------------------------------------*/
int
pickUpCallBack(struct Togl *togl, int argc, char *argv[])
{
	Tcl_Interp *interp = Togl_Interp(togl);
	int x, y, i, j;
	GLuint selectBuf[BUFSIZE];
	char buf[80];
	GLint hits;
	GLuint names, *ptr;
	Tcl_HashEntry *entry;
	GDHE_OBS *obs;
	Tcl_Obj *hit, *result;

	if (argc != 4) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("wrong # args: should be \""
			"ogl pickUp x y\"", -1));
		return TCL_ERROR;
	}
	x = atoi(argv[2]);
	y = atoi(argv[3]);

	/* lookup H-table */
	entry  = Tcl_FindHashEntry(posTable, Togl_Ident(togl));
	if (entry == NULL) {
		Tcl_SetObjResult(interp, 
		    Tcl_NewStringObj("unknown togl ident", -1));
		return TCL_ERROR;
	}
	obs = (GDHE_OBS *)Tcl_GetHashValue(entry);
	obs->renderMode = GL_SELECT;
	obs->pickX = x;
	obs->pickY = y;

	glSelectBuffer(BUFSIZE, selectBuf);
	glRenderMode(GL_SELECT);
	glInitNames();
		
	display_cb(togl);
	glFlush();
	hits = glRenderMode(GL_RENDER);
	obs->renderMode = GL_RENDER;

	result = Tcl_NewObj();
	ptr = (GLuint *)selectBuf;
	for (i = 0; i < hits; i++) {
		hit = Tcl_NewObj();
		names = *ptr++;
		Tcl_ListObjAppendElement(interp, hit, 
		    Tcl_NewDoubleObj((double)*ptr++/(double)UINT32_MAX));
		Tcl_ListObjAppendElement(interp, hit, 
		    Tcl_NewDoubleObj((double)*ptr++/(double)UINT32_MAX));
		for (j = 0; j < names; j++) 
			Tcl_ListObjAppendElement(interp, hit,
			    revNamesArray[*ptr++]);
		Tcl_ListObjAppendElement(interp, result, hit);
	}
	Tcl_SetObjResult(interp, result);
	return TCL_OK;
}
	
/*----------------------------------------------------------------------*/

#ifdef EXEMPLE
/**
 ** interface de 'glXxx'
 **/
int
xxxCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{

    if (argc != 1) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be \""
					       "xxx \"",
					       -1));
	return TCL_ERROR;
    }
    /* glXxx(); */

    /* Let result string empty */
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

#endif /* EXEMPLE */
